/*
 * Name:   goniostat.c
 *
 * Purpose: This is a heavily modified version of the original goniostat.c
 *          that was provided by Michael Blum of Mar-USA.
 *
 * Author:  The original version had the following comments:
 *
 *  goniostat.c - communication program for Spring-8 Beamline 44b2
 *  	      passes command to RAXIS-IV server program running on SGI
 *
 *          The code has been modified by William Lavender to work with MX.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <getopt.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_clock.h"
#include "mx_net.h"
#include "mx_motor.h"
#include "mx_relay.h"
#include "mx_variable.h"
#include "mx_rs232.h"

#define ALLOW_SPEED_CHANGES		FALSE

#define DEFAULT_MX_DATABASE_FILENAME	"/opt/mx/etc/goniostat.dat"
#define EXPOSURE_TYPE_RECORD_NAME	"exposure_type"

static int verbose = 0;
static FILE *LOG = NULL;

extern char *mxgs_fix_record_name( char *record_name );

extern void mxgs_log( char *format, ... ) MX_PRINTFLIKE( 1, 2 );

extern void mxgs_reparse_argv( int old_argc, char **old_argv,
				int *new_argc, char ***new_argv );

extern void mxgs_move_abs_cmd( MX_RECORD *mx_record_list,
				char *motor_name,
				double destination,
				double speed );

extern void mxgs_move_abs_multi_cmd( MX_RECORD *mx_record_list,
					int num_motors,
					char **new_argv );

extern void mxgs_shutter_cmd( MX_RECORD *mx_record_list,
				char *shutter_name,
				int action );

extern void mxgs_query_cmd( MX_RECORD *mx_record_list,
				int num_records,
				char **new_argv,
				int quiet );

extern void mxgs_set_abs_cmd( MX_RECORD *mx_record_list, 
				char *record_name,
				double angle );

extern void mxgs_do_exposure( MX_RECORD *mx_record_list, char *motor_name,
				double angle,
				double seconds,
				int expose );

/* Beamline specific exposure routines. */

extern void mxgs_do_dummy_exposure( MX_RECORD *mx_record_list,
				char *motor_name, double angle,
				double seconds, int expose );

extern void mxgs_do_sercat_exposure( MX_RECORD *mx_record_list,
				char *motor_name, double angle,
				double seconds, int expose );

extern void mxgs_do_adsc_exposure( MX_RECORD *mx_record_list,
				char *motor_name, double angle,
				double seconds, int expose );

int main(int argc, char **argv)
{
	MX_RECORD *mx_record_list;
	char mx_database_filename[MXU_FILENAME_LENGTH+1] = "";
	char logfile_name[MXU_FILENAME_LENGTH+1];
	char program_name[MXU_FILENAME_LENGTH+1];
	char *command_name, *motor_name, *shutter_action;
	double angle, seconds, speed;
	int i, c, error, expose, quiet;
	int new_argc, num_args, num_motors;
	char **new_argv, **argv_ptr;
	mx_status_type mx_status;

	static char usage[] =
	"\n"
	"Usage: %s [-hV] [-f file] [-L file] -- command\n"
	"       -h      print this help message\n"
	"       -V      verbose mode\n"
	"       -q      quiet (for timing tests)\n"
	"       -f file initialize MX from the specified database file\n"
	"       -L file log output to file\n"
	"       --      terminates the list of options\n"
	"\n"
	"  where command is one of:\n"
	"\n"
	"    marccd_move_abs motor_name destination speed [repeats...]\n"
	"\n"
	"    marccd_shutter {open|close}\n"
	"\n"
	"    marccd_query record_name\n"
	"\n"
	"    marccd_set_abs motor_name position (currently disabled)\n"
	"\n"
	"    marccd_exposure motor_name angle_range seconds expose_flag\n"
	"\n";

	/* Process arguments */

	error = FALSE;
	quiet = FALSE;

	while ((c = getopt(argc,argv,"f:hqvVL:")) != EOF) {
		switch(c) {
		case 'V':
			verbose++;
			break;
		case 'L':
			strncpy(logfile_name, optarg, sizeof(logfile_name));
			LOG = fopen(logfile_name, "a");
			break;
		case 'f':
			mx_strncpy( mx_database_filename,
					optarg, MXU_FILENAME_LENGTH );
			break;
		case 'q':
			quiet = TRUE;
			break;
		case 'h':
		case '?':
			error = TRUE;
			break;
		}
	}

	strncpy( program_name, argv[0], sizeof(program_name) );

	if ( error ) {
		mxgs_log( usage, program_name );
		exit(-1);
	}

	if ( (argc - optind) < 1 ) {
		mxgs_log( usage, program_name );
		exit(-1);
	}

	mxgs_reparse_argv( argc - optind, &argv[optind], &new_argc, &new_argv );

	if ( verbose >= 2 ) {
		mxgs_log( "Command line arguments:" );

		for ( i = 0; i < new_argc; i++ ) {
			mxgs_log( " '%s'", new_argv[i] );
		}
		mxgs_log( "\n" );
	}

	if ( strlen( mx_database_filename ) <= 0 ) {
		mx_strncpy( mx_database_filename,
				DEFAULT_MX_DATABASE_FILENAME,
				MXU_FILENAME_LENGTH );
	}

	mx_msleep(1);    /* Initialize millisecond clock support. */

	mx_status = mx_setup_database( &mx_record_list, mx_database_filename );

	if ( mx_status.code != MXE_SUCCESS ) {
		mxgs_log(
	"The attempt to read MX database file '%s' was not successful.\n",
			mx_database_filename );
		exit(-1);
   	}

	command_name = new_argv[0];

	if ( strcmp( command_name, "marccd_move_abs" ) == 0 ) {

		if ( new_argc < 3 ) {
			mxgs_log( usage, program_name );
			exit(-1);
		}

		num_args = new_argc - 1;

		if ( ( num_args % 3 ) != 0 ) {
			mxgs_log(
"Error: Each motor specified by 'marccd_move_abs' should add 3 arguments\n"
"       to the command line, namely, the motor name, the motor destination,\n"
"       and the motor speed.  Thus, the number of command line arguments\n"
"       after 'marccd_move_abs' should be a multiple of 3.  However, the\n"
"       number of arguments you specified was %d which is not a multiple\n"
"       of 3.\n",  num_args );
			exit(-1);
		}

		num_motors = num_args / 3;

		if ( num_motors == 1 ) {
			motor_name = new_argv[1];

			angle = atof( new_argv[2] );

			speed = atof( new_argv[3] );

			mxgs_move_abs_cmd( mx_record_list,
					motor_name, angle, speed );
		} else {
			argv_ptr = new_argv + 1;

			mxgs_move_abs_multi_cmd( mx_record_list,
						num_motors, argv_ptr );
		}

		if ( verbose )
			mxgs_log("Move complete.\n");

	} else if ( strcmp( command_name, "marccd_shutter" ) == 0 ) {

		if ( new_argc < 2 ) {
			mxgs_log( usage, program_name );
			exit(-1);
		}

		shutter_action = new_argv[1];

		if ( strcmp( shutter_action, "open") == 0 ) {

			if ( verbose >= 2 )
				mxgs_log("Opening shutter...\n");

			mxgs_shutter_cmd( mx_record_list,
					"shutter", MXF_OPEN_RELAY );

		} else if ( strcmp( shutter_action, "close") == 0) {

	 		if (verbose >= 2)
				mxgs_log("Closing shutter...\n");

			mxgs_shutter_cmd( mx_record_list,
					"shutter", MXF_CLOSE_RELAY );
		}

	} else if ( strcmp( command_name, "marccd_query" ) == 0 ) {

		if ( new_argc < 2 ) {
			mxgs_log( usage, program_name );
			exit(-1);
		}

		new_argc--;
		new_argv++;

		mxgs_query_cmd( mx_record_list, new_argc, new_argv, quiet );

	} else if ( strcmp( command_name, "marccd_set_abs" ) == 0 ) {

		if ( new_argc < 3 ) {
			mxgs_log( usage, program_name );
			exit(-1);
		}

		motor_name = new_argv[1];

		angle = atof( new_argv[2] );

		mxgs_set_abs_cmd( mx_record_list, motor_name, angle );

	} else if ( strcmp( command_name, "marccd_exposure" ) == 0 ) {

		if ( new_argc < 5 ) {
			mxgs_log( usage, program_name );
			exit(-1);
		}

		motor_name = new_argv[1];

		angle = atof( new_argv[2] );

		seconds = atof( new_argv[3] );

		expose = atoi( new_argv[4] );

		if ( fabs(angle) <= 0.0001 ) {

			/* A tiny angle indicates that this is a still.
			 * We make this explicit by setting the motor
			 * name to a NULL pointer.
			 */

			motor_name = NULL;
			angle = 0.0;
		}

		mxgs_do_exposure( mx_record_list, motor_name,
				angle, seconds, expose );

	} else {
		mxgs_log( "\n%s: Unrecognized command: \"%s\"\n",
				program_name, command_name);

		mxgs_log( usage, program_name);
		exit(-1);
	}
	exit(0);
}

/************/

char *mxgs_fix_record_name( char *record_name )
{
	static char marccd_string[] = "marccd_";
	size_t length;
	char *ptr;

	if ( record_name == NULL ) {
		mxgs_log(
"Error: The record_name pointer passed to 'mxgs_fix_record_name' is NULL.\n" );
		exit(-1);
	}

	ptr = record_name;

	/* If the record name is prefixed with 'marccd_', strip the
	 * prefix off.
	 */

	length = strlen( marccd_string );

	if ( strncmp( marccd_string, ptr, length ) == 0 ) {
		ptr += length;
	}

	/* If the resulting record name is too long, truncate the name. */

	length = strlen( ptr );

	if ( length > MXU_RECORD_NAME_LENGTH ) {
		ptr[ MXU_RECORD_NAME_LENGTH - 1 ] = '\0';
	}

	return ptr;
}

/************/

void mxgs_log( char *format, ... )
{
	va_list args;
	char buffer[2500];

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	fprintf( stderr, "%s", buffer );

	if ( LOG != NULL ) {
		fprintf( LOG, "%s", buffer );
	}
}

/************/

#define ARGV_BLOCK_SIZE    100

void mxgs_reparse_argv( int old_argc, char **old_argv,
			int *new_argc, char ***new_argv )
{
	char **new_argv_array;
	char *ptr;
	int i, j, num_argv_blocks, num_elements;

	/* For some unknown reason, MarCCD concatenates all of its
	 * command line arguments into a single string separated by
	 * spaces and then stuffs that into a single element of
	 * the argv array.  This routine reparses the argv array
	 * creating a new more Unix-like argv array.
	 *
	 * The routine works by leaving the text strings in their
	 * original memory locations and then creating a new array
	 * of row pointers for new_argv that points into the old array
	 * elements.  It also null terminates each string in the old
	 * array elements.
	 *
	 * Note that the routine will also work correctly if passed
	 * normal argv array.
	 */

	if ( old_argc <= 0 ) {
		mxgs_log( "old_argc has the illegal value of %d.\n",
				old_argc );
		mxgs_log( "old_argc must be greater than zero.\n" );
		exit(-1);
	}

	num_argv_blocks = 1;

	num_elements = num_argv_blocks * ARGV_BLOCK_SIZE;

	new_argv_array = (char **) malloc( num_elements );

	if ( new_argv_array == NULL ) {
		mxgs_log(
		"Cannot allocate a %d element array of argv pointers.\n",
				 num_elements );
		exit(-1);
	}

	/*
	 * i = index for old_argv
	 * j = index for new_argv_array
	 */

	j = 0;

	for ( i = 0; i < old_argc; i++ ) {

		ptr = strtok( old_argv[i], " " );

		while ( ptr != NULL ) {
			new_argv_array[j] = ptr;

			j++;

			if ( ( j % ARGV_BLOCK_SIZE ) == 0 ) {

				num_argv_blocks++;

				num_elements =
					num_argv_blocks * ARGV_BLOCK_SIZE;

				new_argv_array = realloc( new_argv_array,
								num_elements );

				if ( new_argv_array == NULL ) {
					mxgs_log(
		"Cannot reallocate a %d element array of argv pointers.\n",
						 num_elements );
					exit(-1);
				}
			}

			ptr = strtok( NULL, " " );
		}
	}

	*new_argc = j;

	*new_argv = new_argv_array;

	return;
}

/************/

void mxgs_move_abs_cmd( MX_RECORD *mx_record_list,
			char *motor_name,
			double destination,
			double speed )
{
	MX_RECORD *motor_record;
	MX_MOTOR *motor;
	int busy, flags;
	mx_status_type mx_status;

	motor_name = mxgs_fix_record_name( motor_name );

	motor_record = mx_get_record( mx_record_list, motor_name );

	if ( motor_record == NULL ) {
		mxgs_log(
	"Cannot find the motor '%s' in the MX database.\n", motor_name );
		exit(-1);
	}

	if ( (motor_record->mx_superclass != MXR_DEVICE)
	  || (motor_record->mx_class != MXC_MOTOR) )
	{
		mxgs_log( "The MX record '%s' is not a motor record.\n",
					motor_record->name );
		exit(-1);
	}

	motor = (MX_MOTOR *) motor_record->record_class_struct;

	if ( motor == (MX_MOTOR *) NULL ) {
		mxgs_log(
		"The MX_MOTOR pointer for motor record '%s' is NULL.\n",
			motor->record->name );
		exit(-1);
	}

#if ALLOW_SPEED_CHANGES

	mx_status = mx_motor_set_speed( motor_record, speed );

	if ( mx_status.code != MXE_SUCCESS ) {
		mxgs_log(
	"Attempt to change the motor speed of '%s' to %f %s/sec failed.\n",
			motor_record->name, speed, motor->units );

		exit(-1);
	}
#endif

	if ( verbose )
		mxgs_log( "Moving '%s' to %f %s.\n",
			motor_record->name,
			destination, motor->units );

	/* Start the move. */

	flags = MXF_MTR_NOWAIT | MXF_MTR_IGNORE_KEYBOARD;

	mx_status = mx_motor_move_absolute( motor_record, destination, flags );

	if ( mx_status.code != MXE_SUCCESS ) {
		mxgs_log( "Attempt to move '%s' to %f %s failed.\n",
			motor_record->name, destination, motor->units );

		exit(-1);
	}

	busy = 1;

	while ( busy != 0 ) {

		mx_status = mx_motor_is_busy( motor_record, &busy );

		if ( mx_status.code != MXE_SUCCESS ) {
			mxgs_log(
		"Attempt to see if motor '%s' is still moving failed.\n",
				motor_record->name );
			exit(-1);
		}
		if ( busy == 0 ) {
			/* Exit the while loop. */

			break;
		}
		mx_msleep(1);
	}

	return;
}

/************/

void mxgs_move_abs_multi_cmd( MX_RECORD *mx_record_list,
					int num_motors,
					char **new_argv )
{
	MX_RECORD **motor_record_array;
	double *destination_array;
	double *speed_array;
	int *still_busy, busy;
	mx_status_type mx_status;
	int i, flags;

#if 0
	for ( i = 0; i < (3 * num_motors); i++ ) {
		mxgs_log("new_argv[%d] = '%s'\n", i, new_argv[i]);
	}
#endif

	/* Attempt to allocate arrays to store the necessary record pointers
	 * and destinations.
	 */

	motor_record_array = (MX_RECORD **)
			malloc( num_motors * sizeof(MX_RECORD *) );

	if ( motor_record_array == (MX_RECORD **) NULL ) {
		mxgs_log( "Error: The attempt to allocate a %d element array "
			"of MX_RECORD pointers failed.\n", num_motors );
		exit(-1);
	}

	destination_array = (double *) malloc( num_motors * sizeof(double) );

	if ( destination_array == (double *) NULL ) {
		mxgs_log( "Error: The attempt to allocate a %d element array "
			"of destination values failed.\n", num_motors );
		exit(-1);
	}

	speed_array = (double *) malloc( num_motors * sizeof(double) );

	if ( speed_array == (double *) NULL ) {
		mxgs_log( "Error: The attempt to allocate a %d element array "
			"of speed values failed.\n", num_motors );
		exit(-1);
	}

	still_busy = (int *) malloc( num_motors * sizeof(int) );

	if ( still_busy == (int *) NULL ) {
		mxgs_log( "Error: The attempt to allocate a %d element array "
			"of busy values failed.\n", num_motors );
		exit(-1);
	}

	for ( i = 0; i < num_motors; i++ ) {

		motor_record_array[i] = mx_get_record( mx_record_list,
							new_argv[ 3*i ] );

		if ( motor_record_array[i] == (MX_RECORD *) NULL ) {
			mxgs_log(
		"Error: The motor name '%s' is not a recognized motor name.\n",
				new_argv[ 3*i ] );
			exit(-1);
		}

		destination_array[i] = atof( new_argv[ 3*i + 1 ] );

		speed_array[i] = atof( new_argv[ 3*i + 2 ] );
	}
#if ALLOW_SPEED_CHANGES

	/* Change the speeds. */

	for ( i = 0; i < num_motors; i++ ) {
		mx_status = mx_motor_set_speed( motor_record_array[i],
							speed_array[i] );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);
	}
#endif

	/* Start the moves. */

	flags = MXF_MTR_NOWAIT | MXF_MTR_IGNORE_KEYBOARD;

	for ( i = 0; i < num_motors; i++ ) {
		mx_status = mx_motor_move_absolute( motor_record_array[i],
							destination_array[i],
							flags );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

		still_busy[i] = TRUE;
	}

	/* Wait for the moves to complete. */

	busy = TRUE;

	while ( busy ) {

		busy = FALSE;

		for ( i = 0; i < num_motors; i++ ) {
			if ( still_busy[i] ) {
				mx_status = mx_motor_is_busy(
							motor_record_array[i],
							&(still_busy[i]) );

				if ( mx_status.code != MXE_SUCCESS )
					exit(-1);

				mx_msleep(1);
			}
			if ( still_busy[i] ) {
				busy = TRUE;
			}
		}
	}

	return;
}

/************/

void mxgs_shutter_cmd( MX_RECORD *mx_record_list,
				char *shutter_name,
				int action )
{
	MX_RECORD *shutter_record;
	mx_status_type mx_status;

	shutter_name = mxgs_fix_record_name( shutter_name );

	shutter_record = mx_get_record( mx_record_list, shutter_name );

	if ( shutter_record == NULL ) {
		mxgs_log(
	"Cannot find the record '%s' in the MX database.\n", shutter_name );
		exit(-1);
	}

	mx_status = mx_relay_command( shutter_record, action );

	if ( mx_status.code != MXE_SUCCESS ) {
		if ( action == MXF_CLOSE_RELAY ) {
			if ( verbose )
				mxgs_log( "Close" );

		} else if ( action == MXF_OPEN_RELAY ) {
			if ( verbose )
				mxgs_log( "Open" );

		} else {
			if ( verbose )
				mxgs_log( "Unrecognized (%d)", action );
		}

		if ( verbose )
			mxgs_log( " command for shutter '%s' failed.\n",
					shutter_record->name );

		exit(-1);
	}
	return;
}

/************/

void mxgs_query_cmd( MX_RECORD *mx_record_list,
			int new_argc, char **new_argv, int quiet )
{
	MX_RECORD *record;
	char *record_name;
	double position;
	int i, relay_status;
	mx_status_type mx_status;

	for ( i = 0; i < new_argc; i++ ) {

		record_name = mxgs_fix_record_name( new_argv[i] );

		record = mx_get_record( mx_record_list, record_name );

		if ( record == (MX_RECORD *) NULL ) {
			mxgs_log(
			"%s is not a valid motor for 'marccd_query'.\n",
				record_name );

			exit(-1);
		}

		if ( strcmp( record_name, "shutter" ) == 0 ) {

			/* 'shutter' is actually a relay record rather
			 * than a motor record, so we have to invoke a
			 * different function.
			 */

			mx_status = mx_get_relay_status(record, &relay_status);

			if ( mx_status.code != MXE_SUCCESS )
				exit(-1);

			/* Convert to Michael's convention. */

			if ( relay_status == 0 ) {
				relay_status = 1;
			} else {
				relay_status = 0;
			}

			if ( quiet == FALSE ) {
				printf( "%d ", relay_status );
			}
		} else {
			mx_status = mx_motor_get_position(record, &position);

			if ( mx_status.code != MXE_SUCCESS )
				exit(-1);

			if ( quiet == FALSE ) {
				printf( "%f ", position );
			}
		}
	}

	printf("\n");

	fflush(stdout);

	return;
}

/************/

void mxgs_set_abs_cmd( MX_RECORD *mx_record_list,
				char *motor_name,
				double angle )
{
	MX_RECORD *motor_record;
	mx_status_type mx_status;

#if 1
	/* Can't allow this to be used until 'marccd_query' is supported
	 * by the MarCCD GUI.  Otherwise, MarCCD will be blindly redefining
	 * motor positions unrelated to the one you _really_ want to 
	 * be redefined.
	 */

	mxgs_log(
	"Error: The 'marccd_set_abs' command is currently disabled.\n" );

	exit(-1);
#endif

	motor_name = mxgs_fix_record_name( motor_name );

	motor_record = mx_get_record( mx_record_list, motor_name );

	if ( motor_record == (MX_RECORD *) NULL ) {
		mxgs_log( "%s is not a valid motor for 'marccd_set_abs'.n",
				motor_name );

		exit(-1);
	}

	mx_status = mx_motor_set_position( motor_record, angle );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	return;
}

/************/

void mxgs_do_exposure( MX_RECORD *mx_record_list,
				char *motor_name,
				double angle,
				double seconds,
				int expose )
{
	MX_CLOCK_TICK start_time, end_time, time_difference;
	double duration;
	char exposure_type[81];
	mx_status_type mx_status;

	mx_initialize_clock_ticks();

	start_time = mx_current_clock_tick();

	mx_status = mx_get_string_variable_by_name( mx_record_list,
					EXPOSURE_TYPE_RECORD_NAME,
				exposure_type, sizeof(exposure_type) - 1 );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	if ( strcmp( exposure_type, "dummy" ) == 0 ) {

		mxgs_do_dummy_exposure( mx_record_list, motor_name,
					angle, seconds, expose );

	} else if ( strcmp( exposure_type, "sercat" ) == 0 ) {

		mxgs_do_sercat_exposure( mx_record_list, motor_name,
					angle, seconds, expose );

	} else if ( strcmp( exposure_type, "adsc" ) == 0 ) {

		mxgs_do_adsc_exposure( mx_record_list, motor_name,
					angle, seconds, expose );

	} else {
		mxgs_log(
"Error: Database variable '%s' is set to an unknown exposure type '%s'\n",
			EXPOSURE_TYPE_RECORD_NAME, exposure_type );
	}

	end_time = mx_current_clock_tick();

	time_difference = mx_subtract_clock_ticks( end_time, start_time );

	duration = mx_convert_clock_ticks_to_seconds( time_difference );

	if ( verbose ) {
		mxgs_log("Exposure complete.  Total time = %g seconds.\n",
			duration);
	}

	return;
}

/************/

void mxgs_do_dummy_exposure( MX_RECORD *mx_record_list,
				char *motor_name,
				double angle,
				double seconds,
				int expose )
{
	unsigned long milliseconds;

	if ( motor_name == NULL ) {
		if ( expose ) {
			mxgs_log( "Exposing: Still for %f seconds...\n",
					seconds );
		} else {
			mxgs_log( "Dark: waiting for %f seconds...\n",
					seconds );
		}
	} else {
		mxgs_log(
		"Exposing: '%s' rotation of %f degrees for %f seconds...\n",
					motor_name, angle, seconds );
	}

	/* Simulate the exposure time for now. */

	milliseconds = mx_round( 1000.0 * seconds );

	mx_msleep( milliseconds );
}

/************/

#define COMMAND_LENGTH	100

void mxgs_do_sercat_exposure( MX_RECORD *mx_record_list,
				char *motor_name,
				double angle,
				double seconds,
				int expose )
{
	static const char fname[] = "mxgs_do_sercat_exposure()";

	char remote_pmac_description[2*MXU_RECORD_NAME_LENGTH + 2];
	char *remote_server_name;
	char *remote_pmac_name;
	MX_RECORD *remote_server_record;
	MX_RECORD *motor_record;

	char command_rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	char response_rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	char command_with_response_rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];

	char command[COMMAND_LENGTH + 1];
	char response[COMMAND_LENGTH + 1];
	double speed, current_position, destination;
	int i, busy, motor_number, num_items;
	unsigned long milliseconds;
	long dimension[1];
	mx_status_type mx_status;

	static struct {
		int number;
		char name[ MXU_RECORD_NAME_LENGTH + 1 ];
	} motor_table[] = {
		{ 1, "omega" },
		{ 2, "phi" },
		{ 3, "kappa" },
		{ 4, "twotheta" },
	};

	static int num_table_entries = sizeof( motor_table )
					/ sizeof( motor_table[0] );

	dimension[0] = COMMAND_LENGTH;

	milliseconds = mx_round( seconds * 1000.0 );

	if ( motor_name != NULL ) {
		motor_record = mx_get_record( mx_record_list, motor_name );

		if ( motor_record == (MX_RECORD *) NULL ) {
			(void) mx_error( MXE_NOT_FOUND, fname,
	"The goniostat database does not contain a motor record named '%s'.",
			motor_name );
			exit(-1);
		}
	}

	mx_status = mx_get_string_variable_by_name( mx_record_list,
					"remote_pmac",
					remote_pmac_description,
					sizeof(remote_pmac_description));

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	remote_server_name = remote_pmac_description;

	remote_pmac_name = strchr( remote_pmac_description, ':' );

	if ( remote_pmac_name == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
	"Did not find the ':' character that separates the remote server name "
	"from the remote pmac name in the variable record 'remote_pmac'" );
		exit(-1);
	}

	*remote_pmac_name = '\0';

	remote_pmac_name++;

	sprintf( command_rfname, "%s.command", remote_pmac_name );

	sprintf( response_rfname, "%s.response", remote_pmac_name );

	sprintf( command_with_response_rfname,
			"%s.command_with_response", remote_pmac_name );

	remote_server_record = mx_get_record( mx_record_list,
						remote_server_name );

	if ( remote_server_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
	"The record 'remote_pmac' says that the PMAC is located in a server "
	"called '%s'.  However, there is no such server defined in "
	"goniostat's database.", remote_server_name );
		exit(-1);
	}

	/* Perform the requested operation. */

	if ( motor_name == NULL ) {

		/* Perform a still. */

		if ( expose ) {
			expose = 1;
		}

		sprintf( command, "&1q40=%dq50=%lub2r", expose, milliseconds );

		mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, command );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

	} else {
		for ( i = 0; i < num_table_entries; i++ ) {
			if ( strcmp( motor_name, motor_table[i].name ) == 0 ) {
				break;	/* Exit the for() loop. */
			}
		}

		if ( i >= num_table_entries ) {
			mxgs_log(
	"Error: '%s' axis exposure is not currently supported at SER-CAT\n",
				motor_name );

			exit(-1);
		}

		motor_number = motor_table[i].number;

		speed = mx_divide_safely( angle, seconds );

		/* Get the current position. */

		mx_status = mx_motor_get_position( motor_record,
							&current_position );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

		mx_info("'%s' position = %g",
			motor_name, current_position );

		/* FIXME: I don't know if Michael Blum wants the angle range
		 * to start at the current position or end at the current
		 * position.  It seems most likely that he wants to start
		 * the frame at the current position, since that would be
		 * more efficient.
		 */

		destination = current_position + angle;

		/* Set parameters for the exposure. */

		sprintf( command, "&1q49=%fq50=%luq%d=%f",
				speed, milliseconds,
				50 + motor_number, destination );
		
		mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, command );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

		if ( verbose ) {
			mxgs_log( "Command = '%s'\n", command );
		}

		/* Do the exposure. */

		sprintf( command, "&1b1r" );

		mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, command );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

		if ( verbose ) {
			mxgs_log( "Exposure started.\n" );
		}
	}

	/* Wait until the motor has stopped moving. */

	while ( busy != 0 ) {

		strcpy( command, "q30" );

		mx_status = mx_put_array( remote_server_record,
					command_with_response_rfname,
					MXFT_STRING, 1, dimension, command );

		if ( mx_status.code != MXE_SUCCESS ) {
			mxgs_log(
		"The '%s' command to see if the exposure is done failed.",
				command );
			exit(-1);
		}

		mx_status = mx_get_array( remote_server_record, response_rfname,
					MXFT_STRING, 1, dimension, response );

		if ( mx_status.code != MXE_SUCCESS ) {
			mxgs_log(
		"Reading the response to the '%s' command failed.", command );
			exit(-1);
		}

		num_items = sscanf( response, "%d", &busy );

		if ( num_items != 1 ) {
			mxgs_log(
	"Attempt to parse the response '%s' to the '%s' command failed.",
				response, command );
			exit(-1);
		}

		if ( busy == 0 ) {
			/* Exit the while loop. */

			break;
		}
		mx_msleep(1);
	}

	return;
}

/************/

#define COMMAND_LENGTH	100

void mxgs_do_adsc_exposure( MX_RECORD *mx_record_list,
				char *motor_name,
				double angle,
				double seconds,
				int expose )
{
	static const char fname[] = "mxgs_do_adsc_exposure()";

	char remote_6k_description[2*MXU_RECORD_NAME_LENGTH + 2];
	char *remote_server_name;
	char *remote_6k_name;
	MX_RECORD *remote_server_record;
	MX_RECORD *motor_record;
	MX_RECORD *rs232_record;

	char command_rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	char response_rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	char command_with_response_rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];

	char command[COMMAND_LENGTH + 1];
	char response[COMMAND_LENGTH + 1];
	double speed, saved_speed;
	double saved_acceleration_params[MX_MOTOR_NUM_ACCELERATION_PARAMS];
	int input_available;
	long dimension[1];
	mx_status_type mx_status;

	char allowed_motor_name[] = "omega";

	dimension[0] = COMMAND_LENGTH;

	/* Get the record corresponding to the motor. */

	if ( motor_name != NULL ) {
		motor_record = mx_get_record( mx_record_list, motor_name );

		if ( motor_record == (MX_RECORD *) NULL ) {
			(void) mx_error( MXE_NOT_FOUND, fname,
	"The goniostat database does not contain a motor record named '%s'.",
				motor_name );
			exit(-1);
		}
	}

	/* Get the record corresponding to the remote RS-232 record. */

	rs232_record = mx_get_record( mx_record_list, "remote_rs232" );

	if ( rs232_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
	"The goniostat database does not contain an RS-232 record named '%s'.",
			"remote_rs232" );
		exit(-1);
	}

	/* Get the information needed to talk to the remote 
	 * Compumotor 6K controller's record.
	 */

	mx_status = mx_get_string_variable_by_name( mx_record_list,
					"remote_6k",
					remote_6k_description,
					sizeof(remote_6k_description));

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	remote_server_name = remote_6k_description;

	remote_6k_name = strchr( remote_6k_description, ':' );

	if ( remote_6k_name == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
	"Did not find the ':' character that separates the remote server name "
	"from the remote compumotor name in the variable record 'remote_6k'" );
		exit(-1);
	}

	*remote_6k_name = '\0';

	remote_6k_name++;

	sprintf( command_rfname, "%s.command", remote_6k_name );

	sprintf( response_rfname, "%s.response", remote_6k_name );

	sprintf( command_with_response_rfname,
			"%s.command_with_response", remote_6k_name );

	remote_server_record = mx_get_record( mx_record_list,
						remote_server_name );

	if ( remote_server_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
"The record 'remote_6k' says that the Compumotor is located in a server "
"called '%s'.  However, there is no such server defined in "
"goniostat's database.", remote_server_name );
		exit(-1);
	}

	/* Perform the requested operation. */

	if ( motor_name == NULL ) {

		/* Perform a still. */

		speed = 0.0;

	} else if ( strcmp( motor_name, allowed_motor_name ) == 0 ) {
		
		/* Save the old speed and acceleration. */

		mx_status = mx_motor_get_speed( motor_record, &saved_speed );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

		mx_status = mx_motor_get_acceleration_parameters(
						motor_record,
						saved_acceleration_params );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

		if ( verbose ) {
			mxgs_log(
			"Saved speed = %f, acceleration = %f, %f, %f, %f\n",
				saved_speed, saved_acceleration_params[0],
				saved_acceleration_params[1],
				saved_acceleration_params[2],
				saved_acceleration_params[3] );
		}

		/* Compute the new speed. */

		speed = mx_divide_safely( angle, seconds );

	} else {
		mxgs_log(
		  "Error: Cannot perform an oscillation with motor '%s'.",
		  motor_name );

		mxgs_log( "Only oscillations using motor '%s' are supported.",
				allowed_motor_name );

		exit(-1);
	}

	if ( verbose ) {
		mxgs_log(
	"Exposure speed = %f, angle range = %f, time = %f\n",
			speed, angle, seconds );
	}

	/* Set the angle range. */

	sprintf( command, "VAR10=%.3f", angle );

	mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, command );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	/* Set the time. */

	sprintf( command, "VAR11=%.3f", seconds );

	mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, command );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	/* Do the exposure. */

	mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, "OSC" );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, "CDONE" );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	if ( verbose ) {
		mxgs_log( "Exposure started.\n" );
	}

	/* Wait until the oscillation program has finished.
	 * It will send a message to the remote RS-232 port
	 * when it has finished.
	 */

	while ( 1 ) {

		/* Read lines of text until we get one that says CMD_DONE. */

		mx_status = mx_rs232_input_is_available( rs232_record,
							&input_available );

		if ( mx_status.code != MXE_SUCCESS )
			exit(-1);

		if ( input_available ) {

			/* Read the line of text sent by the 6K controller. */

			mx_status = mx_rs232_getline( rs232_record,
						response, sizeof(response),
						NULL, 1 );

			if ( mx_status.code != MXE_SUCCESS )
				exit(-1);

			if ( verbose ) {
				mxgs_log("6K response = '%s'\n", response );
			}

			if ( strcmp( response, "CMD_DONE" ) == 0 ) {

				break;   /* Exit the while loop. */
			}
		}
		mx_msleep(1);
	}

	/* Restore the old speeds and accelerations. */

	mx_status = mx_motor_set_speed( motor_record, saved_speed );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	mx_status = mx_motor_set_acceleration_parameters( motor_record,
						saved_acceleration_params );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	/* Put the 6K controller back into absolute positioning mode.  */

	mx_status = mx_put_array( remote_server_record, command_rfname,
					MXFT_STRING, 1, dimension, "MA1" );

	if ( mx_status.code != MXE_SUCCESS )
		exit(-1);

	return;
}

